/****************************************************************************
 *
 * ident1: Beispiel zum Identitaets-Problem
 *
 *  - Beispielprogramm zum Buch "Objektorientiertes Programmieren in C++",
 *    von Nicolai Josuttis, erschienen bei Addison-Wesley, ISBN 3-89319-637-4
 *
 *  - Abschnitt 4.5.3: Das Problem der Identitaet
 *  
 ****************************************************************************/

/****************************************************************************
 *
 * Copyright 1994 by Nicolai Josuttis and Addison-Wesley.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software for personal
 * and educational use is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice
 * and this permission notice appear in supporting documentation, and that
 * the names of Addison-Wesley or the authors are not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.  Addison-Wesley and the authors make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * ADDISON-WESLEY AND THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL ADDISON-WESLEY OR THE AUTHORS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 ****************************************************************************/

// da die meisten Compiler "bool" (noch) nicht koennen (siehe Seite 43):
typedef int bool;
#define true 1
#define false 0



/****************************************************************************
 *
 *	Die Klassenhierarchie
 *
 ****************************************************************************/

#include <iostream.h>


class Fahrzeug {
  private:
    int fahrzeugTeil;
};

class Auto : public Fahrzeug {		// teste auch mit "virtual" !
  private:
    int autoTeil;
};

class Boot : public Fahrzeug {		// teste auch mit "virtual" !
  private:
    int bootTeil;
};

class Amph : public Auto, public Boot {
  private:
    int amphTeil;
};


// Vorwaerts-Deklarationen
void f1 ();
void fAuto (const Auto& a);
void fBoot (const Boot& a);
void fFahrzeug (const Fahrzeug& a);
void f2 ();

main()
{
    cout << "f1():" << endl;
    f1();	// Identitaet bei Referenzen

    cout << "\nf2():" << endl;
    f2();	// Identitaet bei Zeigern
}


/****************************************************************************
 *
 *	Identitaet bei Referenzen:
 *
 ****************************************************************************/

void f1 ()
{
    Amph a;

    fAuto (a);
    fBoot (a);
}

void fAuto (const Auto& a)
{
    cout << "&a  als Auto: " << (void*)&a << endl;
    fFahrzeug (a);
}

void fBoot (const Boot& a)
{
    cout << "&a  als Boot: " << (void*)&a << endl;
    fFahrzeug (a);
}

void fFahrzeug (const Fahrzeug& a)
{
    cout << "    als Fahrzeug: " << (void*)&a << endl;
}


/****************************************************************************
 *
 *	Identitaet bei Zeigern:
 *
 ****************************************************************************/

void f2 ()
{
    Amph a;

    // Adresse von a
    cout << "&a: " << (void*)&a << "\n" << endl;

    // Adresse von a => als Auto, als Boot
    cout << "(Auto*) &a: " << (void*)(Auto*)&a << endl;
    cout << "(Boot*) &a: " << (void*)(Boot*)&a << "\n" << endl;

    // Adresse von a => als Auto => als Fahrzeug
    cout << "(Fahrzeug*) (Auto*) &a: "
         << (void*)(Fahrzeug*)(Auto*)&a << endl;

    // Adresse von a => als Boot => als Fahrzeug
    cout << "(Fahrzeug*) (Boot*) &a: "
         << (void*)(Fahrzeug*)(Boot*)&a << "\n" << endl;
}

